from .exceptions import TransitionError
from .signals import pre_transition_event, post_transition_event
from .utils import parse_receiver_resps
from functools import wraps

import logging
import uuid

logger = logging.getLogger(__name__)


def property_transition(state, from_state, **kwargs):
    return transition(
        state, from_state, to_state=None, is_mutable=False, **kwargs
    )


###############################################################################
#                                                                             #
# TRANSITION DECORATOR                                                        #
#                                                                             #
###############################################################################
def transition(
    state=None, from_state=None, to_state=None, is_mutable=True,
    **decorator_kwargs
):
    """
    transition is model method decorator that flattens the the logic around
    transitioning the state of model. A model attribute is given as the first
    argument and the following two args are the states to transition from and
    to respectively. This is an implementation of a Finite State Machine
    design pattern.
        state => field_name e.g. 'status'
        from_state => state_name or state_number e.g. 'new' or 0
        to_state => state_name or state_number e.g. 'read' or 1
    Usage:
    @transition('read_status', 'new', 'read')
    def read(self):
        # stuff relating to the instance being read
    """
    def decorator(func):
        "First stage of the decorator. It accepts the function to be wraped"
        _function_name = func.__name__
        @wraps(func)
        def _transition(self, *args, **kwargs):
            "wraps the function"
            current_state = getattr(self, state)

            side_effects_enabled = not kwargs.pop('transmission_disable_side_effects', False)

            if current_state == to_state and from_state != to_state:
                # do nothing if the model is already in the desired state
                # but if you want the state to remain the same then continue
                return

            if is_mutable:
                choices = self._meta.get_field(state).choices
                if choices:
                    states = [status for status, _ in choices]
                    if to_state not in states:
                        raise TransitionError(
                            'Cannot transition to a state that is not '
                            'included in the "%s" choices.' % state
                        )
            is_iter = isinstance(from_state, (list, tuple))
            states_align = current_state == from_state
            if (is_iter and current_state in from_state) or states_align:

                thisuuid = str(uuid.uuid4())

                # PRE TRANSITION EVENT
                if side_effects_enabled:
                    receiver_resps = pre_transition_event.send_robust(
                        sender=self.__class__, instance=self, state=state,
                        from_state=current_state, to_state=to_state,
                        transition_name=func.__name__,
                        uuid=thisuuid, **decorator_kwargs
                    )
                    parse_receiver_resps(receiver_resps, log=logger)

                try:
                    # TRANSITION: Run the transition method
                    ret = func(self, *args, **kwargs)
                except:
                    raise
                if side_effects_enabled:
                    if is_mutable:
                        setattr(self, state, to_state)
                    self.save()

                if side_effects_enabled:
                # POST TRANSITION EVENT
                    receiver_resps = post_transition_event.send_robust(
                        sender=self.__class__, instance=self, state=state,
                        from_state=current_state, to_state=to_state,
                        transition_name=func.__name__,
                        uuid=thisuuid, **decorator_kwargs
                    )
                    parse_receiver_resps(receiver_resps, log=logger)

            else:
                raise TransitionError(
                    'Cannot transition from "{current_state!r}"" to '
                    '"{final_state!r}"'.format(
                        current_state=current_state,
                        final_state=to_state or _function_name
                    )
                )
            return ret
        return _transition
    return decorator



from django.db import IntegrityError
from rest_framework.exceptions import APIException, ValidationError


class ServerError(APIException):
    status_code = 500
    default_detail = 'Service malfunction, please try again later.'


class TransitionError(ValidationError, IntegrityError):
    pass



import logging
import sys
import warnings


logger = logging.getLogger(__name__)


def parse_receiver_resps(receiver_resps, log=logger):
    for _, resp in receiver_resps:
        try:
            if isinstance(resp, Exception):
                raise resp
        except Exception, e:
            log.error(unicode(e), exc_info=True)


class RemovedInTransmission2Warning(DeprecationWarning):
    pass


# Deprecated aliases for functions were exposed in this module.

def make_alias(func_name):
    # Close func_name.
    def alias(*args, **kwargs):
        warnings.warn(
            "transmission.utils.%s is deprecated." % func_name,
            RemovedInTransmission2Warning, stacklevel=2
        )
        # This raises a second warning.
        from . import transitions
        return getattr(transitions, func_name)(*args, **kwargs)
    alias.__name__ = func_name
    return alias

this_module = sys.modules['transmission.utils']

for function_name in ('transition', 'property_transition'):
    setattr(this_module, function_name, make_alias(function_name))

del this_module, make_alias, function_name



from django.dispatch import Signal


pre_transition_event = Signal(
    providing_args=[
        "instance", "state", "from_state", "to_state", "transition_name"
    ]
)

post_transition_event = Signal(
    providing_args=[
        "instance", "state", "from_state", "to_state", "transition_name"
    ]
)

pre_event_signal = Signal(
    providing_args=[
        "instance", "event_name", "arg_data", "kwarg_data"
    ]
)

post_event_signal = Signal(
    providing_args=[
        "instance", "event_name", "event_output", "arg_data", "kwarg_data"
    ]
)



from transmission.__version__ import SEMANTIC_VERSION as __version__
from setuptools import setup, find_packages


setup(
    author="hangarunderground",
    author_email="transmission@reelio.com",
    name="transmission",
    packages=find_packages(exclude=['transmission/tests/*']),
    version=__version__,
    url="https://github.com/hangarunderground/transmission",
    download_url=(
        "https://github.com/hangarunderground/transmission/"
        "tarball/" + __version__
    ),
    description=(
        "Transmission is a utility package that houses decorators and "
        "signals pertaining to state transitions and generalized events"
    ),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    keywords=["django", "finite state machine", "rest", "signals"],
    install_requires=['django', 'djangorestframework'],
    extras_require={
        'dev': ['ipdb', 'mock', 'funcsigs', 'pylint', 'Fabric', 'tox'],
    }
)



from .exceptions import ServerError
from .signals import pre_event_signal, post_event_signal
from django.db import transaction
from functools import wraps

import logging
import sys


logger = logging.getLogger(__name__)


def send_event_signals(func):
    """
        `send_event_signals` is an instance method decorator that is designed
        to generate pre and post signals for the method that it is
        decoratoring.
        E.g.
        class SomeModel(models.Model):
            @send_event_signals
            def say_hello(self):
                ...

        will send a pre_event_signal and a post_event_signal signal that will
        have the event_name="say_hello"
    """
    @wraps(func)
    def _wrap(self, *args, **kwargs):
        try:
            with transaction.atomic():
                # PRE EVENT SIGNAL
                pre_event_signal.send(
                    sender=self.__class__, instance=self,
                    event_name=func.__name__, arg_data=args, kwarg_data=kwargs
                )

                # EXECUTE ORIGINAL FUNCTION
                output = func(self, *args, **kwargs)

                # POST EVENT SIGNAL
                post_event_signal.send(
                    sender=self.__class__, instance=self,
                    event_name=func.__name__, event_output=output,
                    arg_data=args, kwarg_data=kwargs
                )

                return output
        except Exception, e:
            logger.error(unicode(e), exc_info=True)
            exc = type('EventError', (ServerError, e.__class__), {})
            raise exc(
                detail=(
                    'Sorry there was an error in processing '
                    'your request, please try again later.'
                )
            ), None, sys.exc_info()[2]
    return _wrap
